//*****************************************************************
// Questo libreria si occupa di implementare le funzioni di una 
// ALU all'interno di un una macchina tipo quella di Von Neumann 
//
// Nome del progetto     : alu_unit.h alu_unit.cpp
// Creatore              : Giovanni Ceglia
// Data ultima revisione : 16/05/2001
//
// Librerie  richieste: vmachine.h vmachine.cpp
//*****************************************************************
 
#include <iostream.h>
#include "vmachine.h"

// funzione che gestisce le istruzioni aritmetico logiche.
// questa funzione opera su una struttura astratta che 
// rappresenta una macchina virtuale.
void alu_exec(virtual_machine&);

// funzione per l'incremento 
void i_inc(virtual_machine& vm);

// funzione per il decremento
void i_dec(virtual_machine& vm);

// funzione per il confronto di due valori
void i_cmp(virtual_machine& vm);

// funzione per l'addizione
void i_add(virtual_machine& vm);

// funzione per la sottrazione
void i_sub(virtual_machine& vm);

// funzione per la moltiplicazione
void i_mul(virtual_machine& vm);

// funzione per la divisione
void i_div(virtual_machine& vm);

// incrementa il valore nella cella di memoria puntata dal registro A
void i_inc(virtual_machine& vm)
 {
  char val;
  val=vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset];
  if (int(val)==127) {
  
    vm.virtual_flags=(vm.virtual_flags | 0x04); // SET Flag Overflow To 1
    val=0;
   } else val++;
  if (int(val)==0) vm.virtual_flags=(vm.virtual_flags | 0x02); // SET Flag Zero To 1
  if (int(val)>0) vm.virtual_flags=(vm.virtual_flags | 0x08); // SET Flag Positive To 1
  if (int(val)<0) vm.virtual_flags=(vm.virtual_flags | 0x10); // SET Flag Negative To 1
  vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset]=val;
 }

// decrementa il valore nella cella di memoria puntata dal registro A
void i_dec(virtual_machine& vm)
 {
  char val;
  val=vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset];
  if (int(val)==-128) {
    vm.virtual_flags=(vm.virtual_flags | 0x04); // SET Flag Overflow To 1
    val=0;
   } else val--;
  if (int(val)==0) vm.virtual_flags=(vm.virtual_flags | 0x02); // SET Flag Zero To 1
  if (int(val)>0) vm.virtual_flags=(vm.virtual_flags | 0x08); // SET Flag Positive To 1
  if (int(val)<0) vm.virtual_flags=(vm.virtual_flags | 0x10); // SET Flag Negative To 1
  vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset]=val;
 }

// effettua un confronto fra due valori puntati dai due operandi A e B
// a seconda dei valori aggiorna i flags di stato.
void i_cmp(virtual_machine& vm)
 {
  if (vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset]==vm.virtual_memory[vm.reg_opcode_B])
     vm.virtual_flags=(vm.virtual_flags | 0x02); // SET Flag Zero To 1
   else if (vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset]>vm.virtual_memory[vm.reg_opcode_B])
     vm.virtual_flags=(vm.virtual_flags | 0x08); // SET Flag Positive To 1
    else vm.virtual_flags=(vm.virtual_flags | 0x10); // SET Flag Negative To 1
 }


// effettua la addizione fra due valori puntati dai due operandi A e B
// il risulatato va in  A
void i_add(virtual_machine& vm)
 {
  char v1,v2;
  v1=vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset];
  v2=vm.virtual_memory[vm.reg_opcode_B];
  if ((int(v1)+int(v2))>127) {
     vm.virtual_flags=(vm.virtual_flags | 0x04); // SET Flag Overflow To 1
     v1=0;
   } else v1=v1+v2;
  vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset]=v1;
  if (v1==0) vm.virtual_flags=(vm.virtual_flags | 0x02); // SET Flag Zero To 1
  if (v1>0) vm.virtual_flags=(vm.virtual_flags | 0x08); // SET Flag Positive To 1
  if (v1<0) vm.virtual_flags=(vm.virtual_flags | 0x10); // SET Flag Negative To 1
 }

// effettua la sottrazione fra due valori puntati dai due operandi A e B
// il risulatato va in  A
void i_sub(virtual_machine& vm)
 {
  char v1,v2;
  v1=vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset];
  v2=vm.virtual_memory[vm.reg_opcode_B];
  if ((int(v1)-int(v2))<-128) {
     vm.virtual_flags=(vm.virtual_flags | 0x04); // SET Flag Overflow To 1
     v1=0;
   } else v1=v1-v2;
  vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset]=v1;
  if (v1==0) vm.virtual_flags=(vm.virtual_flags | 0x02); // SET Flag Zero To 1
  if (v1>0) vm.virtual_flags=(vm.virtual_flags | 0x08); // SET Flag Positive To 1
  if (v1<0) vm.virtual_flags=(vm.virtual_flags | 0x10); // SET Flag Negative To 1
 }

// effettua la moltiplicazione fra due valori puntati dai due operandi A e B
// il risulatato va in  A
void i_mul(virtual_machine& vm)
 {
  char v1,v2;
  v1=vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset];
  v2=vm.virtual_memory[vm.reg_opcode_B];
  if ((int(v1)*int(v2))>127) {
     vm.virtual_flags=(vm.virtual_flags | 0x04); // SET Flag Overflow To 1
     v1=0;
   } else v1=v1*v2;
  vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset]=v1;
  if (v1==0) vm.virtual_flags=(vm.virtual_flags | 0x02); // SET Flag Zero To 1
  if (v1>0) vm.virtual_flags=(vm.virtual_flags | 0x08); // SET Flag Positive To 1
  if (v1<0) vm.virtual_flags=(vm.virtual_flags | 0x10); // SET Flag Negative To 1
 }

// effettua una divisione fra valori contenuti in due indirizzi della macchina
// virtuale e restituisce in uno il risultato della divisione e nell'altro 
// il resto.
void i_div(virtual_machine& vm)
 {
  char v1,v2;
  v1=vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset];
  v2=vm.virtual_memory[vm.reg_opcode_B];
  if (v2==0) (vm.virtual_flags | 0x20); // SET Flag dIv_X_ZeRo To 1
   else {
      vm.virtual_memory[vm.reg_opcode_A+vm.reg_offset]=int(v1) / int(v2);
      vm.virtual_memory[vm.reg_opcode_B]=int(v1) % int(v2);
      if (v1==0) vm.virtual_flags=(vm.virtual_flags | 0x02); // SET Flag Zero To 1
      if (v1>0) vm.virtual_flags=(vm.virtual_flags | 0x08); // SET Flag Positive To 1
      if (v1<0) vm.virtual_flags=(vm.virtual_flags | 0x10); // SET Flag Negative To 1
    }
 }

// funzione che gestisce le istruzioni aritmetico logiche.
// questa funzione opera su una struttura astratta che 
// rappresenta una macchina virtuale.

void alu_exec(virtual_machine& vm)
 {
   switch (vm.reg_istruzione) {
      // operazioni destinate all'unit ALU (Unit Logico Aritmetica)
      case INC : i_inc(vm); break;
      case DEC : i_dec(vm); break;
      case CMP : i_cmp(vm); break;
      case ADD : i_add(vm); break;
      case SUB : i_sub(vm); break;
      case MUL : i_mul(vm); break;
      case DIV : i_div(vm); break;
      case INCOFFST: vm.reg_offset++; break;
      case DECOFFST: vm.reg_offset--; break;
      default  : break;
    }
 } 
 

