//*****************************************************************
// Questa libreria definisce ed implementa le funzioni utili
// alla gestione della struttura astratta che rappresenta la 
// macchina virtuale da simulare.
//
// Nome del progetto     : vmachine.h vmachine.cpp
// Creatore              : Giovanni Ceglia
// Data ultima revisione : 22/05/2003
//
// richiede iostrem.h, conio.h stdio.h
// inoltre necessita del modulo tplus.h e tplus.cpp
//*****************************************************************

#include <iostream.h>
#include <conio.h>
#include <stdio.h>
#include "tplus.h"

// definizione dei tipi di istruzioni e del nome e del codice 
// operativo associato ad esse.

// Istruzioni con nessun operando. 

#define RST 0    // Istruzione di Reset, interrompe la simulazione.
#define STD 1    // Setta a 1 il flag di Direzione.
#define STZ 2    // Setta a 1 il flag di Zero.
#define STO 3    // Setta a 1 il flag di Overflow. 
#define STP 4    // Setta a 1 il flag di Positivit
#define STN 5    // Setta a 1 il flag di Negativit
#define NOP 6    // Non esegue nessuna operazione.
#define CLD 7    // Cancella e mette a 0 il flag Direzione.
#define CLZ 8    // Cancella e mette a 0 il flag Zero.
#define CLO 9    // Cancella e mette a 0 il flag Overflow.
#define CLP 10   // Cancella e mette a 0 il flag Positivit
#define CLN 11   // Cancella e mette a 0 il flag Negativit

#define INCOFFST 12 // Incrementa il registro di Offset, il quale
                    //  un registro il cui valore viene sommato
                    // in qualsiasi operazione di accesso alla 
                    // memoria.
                    
#define DECOFFST 13 // Decrementa il registro di Offset.

// Istruzioni con 1 solo operando.

#define JMP 64   // Salto incondizionato a 8 bit. JMP address
#define JPD 65   // Salta se Direzione = 1        JPD address 
#define JPO 66   // Salta se Overflow = 1         JPO address
#define JPZ 67   // Salta se Zero = 1             JPZ address
#define JPP 68   // Salta se Positivo = 1         JPP address
#define JPN 69   // Salta se Negativo = 1         JPN address

#define INP 70   // Legge Stringa da Tastiera     INP address
#define WRT 71   // Scrive Stringa a Video        WRT address
#define INC 72   // Incrementa una cella di mem.  INC address
#define DEC 73   // Decrementa una cella di mem.  DEC address

#define VALOFFST 74 // Assegna un valore al registro di OffSet.

// Istruzioni con due operandi.

#define CMP 128  // Compara due celle di memoria CMP address1,address2
                 // se i valori sono uguali setta il flag ZERO = 1
                 // se address1<address2 setta il flag Negativo = 1
                 // se address1>address2 setta Positivo = 1 
#define ADD 129  // Somma         ADD addr1,addr2  --> addr1=addr1+addr2
#define SUB 130  // Sottrazione   SUB addr1,addr2  --> addr1=addr1-addr2
#define MUL 131  // Moltiplica    MUL addr1,addr2  --> addr1=addr1*addr2
#define DIV 132  // Divisione     DIV addr1,addr2  --> addr1=addr1/addr2
#define CPY 133  // Copia         CPY addr1,addr2  --> addr1=addr2
#define VAL 134  // Assegna Val.  VAL addr1,valore --> addr1=valore                


// definizione della struttura astratta  che rappresenta una macchina
// virtuale e che mantiene tutte le informazioni inerenti al suo avanzamento
// e funzionamento
struct virtual_machine {
  unsigned char virtual_flags;
  unsigned char virtual_memory[256];
  unsigned char virtual_PI;
  unsigned char reg_istruzione;
  unsigned char reg_opcode_A;
  unsigned char reg_opcode_B;
  unsigned char reg_offset;
 };

// definizione di un tipo che rappresenta il codice operativo 
// dell'istruzione.
typedef unsigned char lowcode;

// restituisce il codice operativo dell'istruzione, 
// conoscendone la rappresentazione come stringa.
lowcode istructioncode(const string81 s);

// carica un programma codificato in codice macchina, e lo 
// trasferisce nella memoria della macchina virtuale, ed inoltre
// inizializza la macchina per il suo avvio.
void carica_programma(const string81, virtual_machine&);

// funzione che  attivata quando la macchina ha il flag di debug
// attivo, e che permette di visualizzare lo stato della macchina.
void visualizza_stato_macchina(const virtual_machine&);

// visualizza il codice mnemonico di una istruzzione, conoscendo
// il codice operativo.
void decode_i(const lowcode code);

// implementazioni delle funzioni e dei sottoprogrammi sopra descritti.

lowcode istructioncode(const string81 s)
 {
   if (ceglia_strcmp(s,"NOP")==true) return NOP;
    else if (ceglia_strcmp(s,"RST")==true) return RST;
    else if (ceglia_strcmp(s,"STD")==true) return STD;
    else if (ceglia_strcmp(s,"STZ")==true) return STZ;
    else if (ceglia_strcmp(s,"STO")==true) return STO;
    else if (ceglia_strcmp(s,"STP")==true) return STP;
    else if (ceglia_strcmp(s,"STN")==true) return STN;
    else if (ceglia_strcmp(s,"CLD")==true) return CLD;
    else if (ceglia_strcmp(s,"CLZ")==true) return CLZ;
    else if (ceglia_strcmp(s,"CLO")==true) return CLO;
    else if (ceglia_strcmp(s,"CLP")==true) return CLP;
    else if (ceglia_strcmp(s,"CLN")==true) return CLN;
    else if (ceglia_strcmp(s,"JMP")==true) return JMP;
    else if (ceglia_strcmp(s,"JPD")==true) return JPD;
    else if (ceglia_strcmp(s,"JPZ")==true) return JPZ;
    else if (ceglia_strcmp(s,"JPO")==true) return JPO;
    else if (ceglia_strcmp(s,"JPP")==true) return JPP;
    else if (ceglia_strcmp(s,"JPN")==true) return JPN;
    else if (ceglia_strcmp(s,"INP")==true) return INP;
    else if (ceglia_strcmp(s,"WRT")==true) return WRT;
    else if (ceglia_strcmp(s,"INC")==true) return INC;
    else if (ceglia_strcmp(s,"DEC")==true) return DEC;
    else if (ceglia_strcmp(s,"CMP")==true) return CMP;
    else if (ceglia_strcmp(s,"ADD")==true) return ADD;
    else if (ceglia_strcmp(s,"SUB")==true) return SUB;
    else if (ceglia_strcmp(s,"MUL")==true) return MUL;
    else if (ceglia_strcmp(s,"DIV")==true) return DIV;
    else if (ceglia_strcmp(s,"CPY")==true) return CPY;
    else if (ceglia_strcmp(s,"VAL")==true) return VAL;
    else if (ceglia_strcmp(s,"VALOFFST")==true) return VALOFFST;
    else if (ceglia_strcmp(s,"INCOFFST")==true) return VALOFFST;
    else if (ceglia_strcmp(s,"DECOFFST")==true) return VALOFFST;
     else return NOP;
 }

void visualizza_stato_macchina(const virtual_machine& vm)
 {
     cout<<"\n";
     cout<<"----------------------------- Virtual Processor State ------------------------"<<"\n";
     cout<<"PROSSIMA ISTRUZIONE         = "<<int(vm.virtual_PI)<<"\n";
     cout<<"REGISTRO DEI FLAGS          = ";
     if ((vm.virtual_flags & 0x01)>0) cout<<"Debug ";
     if ((vm.virtual_flags & 0x02)>0) cout<<"Zero ";
     if ((vm.virtual_flags & 0x04)>0) cout<<"Overflow ";
     if ((vm.virtual_flags & 0x08)>0) cout<<"Positive ";
     if ((vm.virtual_flags & 0x10)>0) cout<<"Negative ";
     if ((vm.virtual_flags & 0x20)>0) cout<<"dIv_x_ZeRo!!! ";
     cout<<"\n";
     cout<<"ISTRUZIONE DA ESEGUIRE      = ";
     decode_i(vm.reg_istruzione);
     cout<<"\n";
     cout<<"OPERANDO A                  = "<<int(vm.reg_opcode_A)<<"\n";
     cout<<"OPERANDO B                  = "<<int(vm.reg_opcode_B)<<"\n";
     cout<<"REGISTRO DI SPIAZZAMENTO    = "<<int(vm.reg_offset)<<"\n";
     cout<<"------------------------------------------------------------------------------"<<"\n";
     cout<<"\n";
     getche();
 }

void decode_i(const lowcode code)
 {
  switch (code) {
    case RST : cout<<"RST"; break;
    case STD : cout<<"STD"; break;
    case STZ : cout<<"STZ"; break;
    case STO : cout<<"STO"; break;
    case STP : cout<<"STP"; break;
    case STN : cout<<"STN"; break;
    case NOP : cout<<"NOP"; break;
    case CLD : cout<<"CLD"; break;
    case CLZ : cout<<"CLZ"; break;
    case CLO : cout<<"CLO"; break;
    case CLP : cout<<"CLP"; break;
    case CLN : cout<<"CLN"; break;
    case JMP : cout<<"JMP"; break;
    case JPD : cout<<"JPD"; break;
    case JPZ : cout<<"JPZ"; break;
    case JPO : cout<<"JPO"; break;
    case JPP : cout<<"JPP"; break;
    case JPN : cout<<"JPN"; break;
    case INP : cout<<"INP"; break;
    case WRT : cout<<"WRT"; break;
    case INC : cout<<"INC"; break;
    case DEC : cout<<"DEC"; break;
    case CMP : cout<<"CMP"; break;
    case ADD : cout<<"ADD"; break;
    case SUB : cout<<"SUB"; break;
    case MUL : cout<<"MUL"; break;
    case DIV : cout<<"DIV"; break;
    case CPY : cout<<"CPY"; break;
    case VAL : cout<<"VAL"; break;
    case VALOFFST : cout<<"VALOFFST"; break;
    case INCOFFST : cout<<"VALOFFST"; break;
    case DECOFFST : cout<<"VALOFFST"; break;
    default  : cout<<"NOP"; break;
   } 
 }


void carica_programma(const string81 nomeprogramma, virtual_machine& vm)
 {
   vm.virtual_flags=0;
   vm.virtual_PI=0;
   vm.reg_offset=0;
   vm.reg_opcode_A=0;
   vm.reg_opcode_B=0;
   FILE *f;
   if ((f = fopen(nomeprogramma, "rb")) != NULL)
    {
      unsigned int i=0;
      while (!feof(f))
       {
         vm.virtual_memory[i]=getc(f);
         i++;
       }
     fclose(f);
    }
 }

